==> org-lsst-ccs-subsystem-archon-main-2.0.1-sources.jar: org/lsst/ccs/subsystem/archon/FitsUtilities.java <==
package org.lsst.ccs.subsystem.archon;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.StandardCopyOption;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;
import nom.tam.fits.BasicHDU;
import nom.tam.fits.BinaryTable;
import nom.tam.fits.BinaryTableHDU;
import nom.tam.fits.Fits;
import nom.tam.fits.FitsException;
import nom.tam.fits.Header;
import nom.tam.util.BufferedFile;
import org.lsst.ccs.utilities.image.FitsCheckSum;

/**
 * Some utilities for dealing with test stand fits files.
 *
 * @author homer
 * @author tonyj
 */
public class FitsUtilities {

    static class PDValue {

        private double time;
        private double value;

        private PDValue(double time, double value) {
            this.time = time;
            this.value = value;
        }

        public double getTime() {
            return time;
        }

        public double getValue() {
            return value;
        }
    }

    List<PDValue> readPhotoDiodeFile(File fileName) throws FileNotFoundException, IOException {
        try (BufferedReader in = new BufferedReader(new FileReader(fileName))) {
            List<PDValue> result = new ArrayList<>();
            for (;;) {
                String line = in.readLine();
                if (line == null) {
                    break;
                }
                String[] tokens = line.split(" ");
                double time = Double.parseDouble(tokens[0]);
                double value = Double.parseDouble(tokens[1]);
                result.add(new PDValue(time, value));
            }
            return result;
        }
    }

    // this returns the integration of PD(t)*dt over the shutter open period 
    double analyzePhotoDiodeValues(List<PDValue> pddata) {
        final double FRAC_SIG_MIN_LIMIT = 0.05;
        double pdsum = 0.0, pdsum_sig = 0.0, pdsum_base = 0.0;
        int nelem_sig = 0, nelem_base = 0;
// get average of all values
        for (PDValue data : pddata) {
            pdsum += data.getValue();
        }
        double pdavg_all = pdsum / pddata.size();
        System.out.println("Average of all PD values is " + pdavg_all);

// now get high and low levels
        pdsum_sig = 0.0;
        for (PDValue data : pddata) {
            double pdval = data.getValue();
            if (pdval < pdavg_all) {
                pdsum_sig += pdval;
                nelem_sig++;
            } else {
                pdsum_base += pdval;
                nelem_base++;
            }
        }
        double pdavg_sig = pdsum_sig  / nelem_sig;
        double pdavg_base = pdsum_base / nelem_base;
        System.out.println("Number elements in shutter open region = " + nelem_sig);
        // now integrate points in shutter open range
        double pdval_previous = 0.,
               pdtime_previous = 0.,
               pdtime_sum = 0.,
               frac_sig_previous = 0.;
        int    nint = 0;
        double pdint = 0.0; // SUM PD(t)*dt

        for (PDValue data : pddata) {
            double pdval = data.getValue();
            double pdtime = data.getTime();
            double frac_sig = (pdval-pdavg_base)/(pdavg_sig-pdavg_base);
            if (frac_sig>FRAC_SIG_MIN_LIMIT || frac_sig_previous>FRAC_SIG_MIN_LIMIT) { // use value if it or its previous value is at least FRAC_SIG_MIN_LIMIT sig level
                double delta_time = pdtime-pdtime_previous;
                // note that the following is corrected for a non-zero baseline
                pdint += ((pdval+pdval_previous)/2.0-pdavg_base)*delta_time;
                pdtime_sum += delta_time;
                nint++;
            }
            pdval_previous    = pdval;
            pdtime_previous   = pdtime;
            frac_sig_previous = frac_sig;
        }
                
        double pdavg_expo = pdint / pdtime_sum;

        System.out.println("Sum of PD values in luminous region = " + pdsum);
        System.out.println("PD average value during exposure = " + pdavg_sig);
        System.out.println("PD average value baseline = " + pdavg_base);
        System.out.println("PD average value sig-baseline = " + (pdavg_sig - pdavg_base));
        System.out.println("Number elements integrated = " + nint);
        System.out.println("Integration summed time = " + pdtime_sum);
        System.out.println("PD average from integration = " + pdavg_expo);
        System.out.println("NOTE: The MONDIODE value will use EXPTIME.");
        return pdint / 1.e-9; // convert to nA
    }

    private BinaryTableHDU createPhotoDiodeHDU(List<PDValue> data,String extnam, String c1name, String c2name, double tstart) throws FitsException {
        BinaryTable binTable = new BinaryTable();
        double[] times = new double[data.size()];
        double[] values = new double[data.size()];
        int i = 0;
        for (PDValue datum : data) {
            times[i] = datum.getTime();
            values[i] = datum.getValue();
            i++;
        }
        binTable.addColumn(times);
        binTable.addColumn(values);

        Header header = new Header(binTable);
        
        header.addValue("EXTNAME", extnam,"Name of the extension");
        header.addValue("TSTART", tstart, "Time of Start of Readings");

        final BinaryTableHDU bhdu = new BinaryTableHDU(header, binTable);
        bhdu.setColumnName(0, c1name, "time");
        bhdu.setColumnName(1, c2name, "values");
        bhdu.info(System.out);
        return bhdu;
    }

    /**
     * Update an existing files by adding or replacing the extension HDU
     * containing the binary table of photo diode values. Note this routine
     * currently works by copying the entire file to a temporary file, then
     * renaming the temporary file over the top of the original file.
     *
     * @param pddatfile The file containing the photo diode data
     * @param fitsfile The fits file to update
     * @param extnam The extension name to add/update
     * @param c1name The name for the first column of data
     * @param c2name The name for the second column of data
     * @param tstart The start time of the readings?
     * @throws java.io.IOException
     * @throws nom.tam.fits.FitsException
     */
    public void updatePhotoDiodeValues(File pddatfile, File fitsfile, String extnam, String c1name, String c2name, double tstart) throws IOException, FitsException {

        List<PDValue> data = readPhotoDiodeFile(pddatfile);
        double pdint = analyzePhotoDiodeValues(data);
        double pdavg = 0.; // guard value
        BinaryTableHDU bhdu = createPhotoDiodeHDU(data,extnam,c1name,c2name,tstart);

        File tmpName = File.createTempFile("tmp", "fits", fitsfile.getParentFile());
        try (Fits in = new Fits(fitsfile)) {
            try (BufferedFile out = new BufferedFile(tmpName, "rw")) {
                in.read();
                BasicHDU primary = in.getHDU(0);
                double exptime = primary.getHeader().getDoubleValue("EXPTIME",1.0);
                if (primary.getHeader().containsKey("MONDIODE")) {
                    primary.getHeader().deleteKey("MONDIODE");
                }
                pdavg = 0.;
                if (exptime>0.0) pdavg = pdint / exptime;
                primary.getHeader().addValue("MONDIODE", pdavg, "avg PD value (nA) during expo");
                FitsCheckSum.setChecksum(primary);
                primary.write(out);
                for (int ihdu = 1; ihdu < in.getNumberOfHDUs(); ihdu++) {
                    final BasicHDU hdu = in.getHDU(ihdu);
                    String extn = hdu.getHeader().getStringValue("EXTNAME");
                    if (!extn.equals(extnam)) {
                        hdu.write(out);
                    }
                }
                FitsCheckSum.setChecksum(bhdu);
                bhdu.write(out);
            }
        }
        Files.move(tmpName.toPath(), fitsfile.toPath(), StandardCopyOption.REPLACE_EXISTING);
    }

    /**
     * Compute the flux statistics by reading an existing FITS file
     *
     * @param fitsfile The file to read
     * @return The computed flux statistics.
     * @throws nom.tam.fits.FitsException If the fits file can not be read
     * @throws java.io.IOException If an IO exception occurs
     */
    public double getFluxStats(File fitsfile) throws FitsException, IOException {

        try (Fits fits = new Fits(fitsfile)) {
            double minflux = Double.MAX_VALUE;
            double fluxsum = 0.0;
                double[] fluxvals = new double[16];
            int fluxentries = 0;
            
            fits.read();
            final Header primary = fits.getHDU(0).getHeader();
            double gain = primary.getDoubleValue("CCDGAIN");
            double exptime = fits.getHDU(0).getHeader().getDoubleValue("EXPTIME");

            for (int ihdu = 1; ihdu < fits.getNumberOfHDUs(); ihdu++) {
                final Header extension = fits.getHDU(ihdu).getHeader();
                String extn = extension.getStringValue("EXTNAME");
                if (extn.contains("Segment")) {
                    double avg = fits.getHDU(ihdu).getHeader().getDoubleValue("AVERAGE"); // data region only
                    double bias = fits.getHDU(ihdu).getHeader().getDoubleValue("AVGBIAS"); // bias region only
                    double signal = (avg - bias) * gain;         // in electrons
                    double flux = Math.max(signal / exptime, 1.0);  // in electrons/second (can't be < 0 !)
                    minflux = Math.min(minflux, flux);
                    fluxsum += flux;
                    fluxvals[fluxentries] = flux;
                    fluxentries++;
                    System.out.println(extn + " signal = " + signal + " exptime = " + exptime + " flux = " + flux + " minflux for CCD = " + minflux);
                }
            }

            // ignore bad segments by only using flux values that are no less than a third of the average
            double flux_cutoff = fluxsum / (3.0 * fluxentries);

            int num_used = 0;
            double flux_resum = 0.0;
            for (int idx=0; idx<16 ; idx++) {
                if (fluxvals[idx]>flux_cutoff) {
                    flux_resum += fluxvals[idx];
                    num_used++;
                }
            }
            System.out.println("flux_cutoff = "+flux_cutoff+" , number flux values used = "+num_used);
            double sensor_flux = num_used > 0 ? flux_resum / num_used : 0.0 ;
//            double sensor_flux = (fluxsum - minflux) / (fluxentries-1);
            System.out.println("flux with all segments = "+fluxsum/fluxentries);
            System.out.println("flux with non-responsive segments removed = "+sensor_flux);
//            System.out.println("flux with min value removed = "+sensor_flux);
            return sensor_flux;
        }
    }
}

==> org-lsst-ccs-subsystem-archon-main-2.0.3-sources.jar: org/lsst/ccs/subsystem/archon/FitsUtilities.java <==
package org.lsst.ccs.subsystem.archon;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.StandardCopyOption;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;
import nom.tam.fits.BasicHDU;
import nom.tam.fits.BinaryTable;
import nom.tam.fits.BinaryTableHDU;
import nom.tam.fits.Fits;
import nom.tam.fits.FitsException;
import nom.tam.fits.Header;
import nom.tam.util.BufferedFile;
import org.lsst.ccs.utilities.image.FitsCheckSum;

/**
 * Some utilities for dealing with test stand fits files.
 *
 * @author homer
 * @author tonyj
 */
public class FitsUtilities {

    static class PDValue {

        private double time;
        private double value;

        private PDValue(double time, double value) {
            this.time = time;
            this.value = value;
        }

        public double getTime() {
            return time;
        }

        public double getValue() {
            return value;
        }
    }

    List<PDValue> readPhotoDiodeFile(File fileName) throws FileNotFoundException, IOException {
        try (BufferedReader in = new BufferedReader(new FileReader(fileName))) {
            List<PDValue> result = new ArrayList<>();
            for (;;) {
                String line = in.readLine();
                if (line == null) {
                    break;
                }
                String[] tokens = line.split(" ");
                double time = Double.parseDouble(tokens[0]);
                double value = Double.parseDouble(tokens[1]);
                result.add(new PDValue(time, value));
            }
            return result;
        }
    }

    // this returns the integration of PD(t)*dt over the shutter open period 
    double analyzePhotoDiodeValues(List<PDValue> pddata) {
        final double FRAC_SIG_MIN_LIMIT = 0.05;
        double pdsum = 0.0, pdsum_sig = 0.0, pdsum_base = 0.0;
        double GUARD = 999.;
        double pdmax = -GUARD, pdmin = GUARD;
        int nelem_sig = 0, nelem_base = 0;
// get average of all values
        for (PDValue data : pddata) {
            double pdval = data.getValue();
            pdsum += pdval;
            if (pdval != 0.0) {
                if (pdval < pdmin) pdmin = pdval;
                if (pdval > pdmax) pdmax = pdval;
            }
        }
        double pdavg_all = pdsum / pddata.size();
        System.out.println("Average of all PD values is " + pdavg_all);

//        double pdcut = pdavg_all ; // leave this as a historical ref
        double pdcut = (pdmin + pdmax) / 2.0;
        System.out.println("The cut between the two PD levels is " + pdcut);
        
// now get high and low levels
        pdsum_sig = 0.0;
        for (PDValue data : pddata) {
            double pdval = data.getValue();
            if (pdval < pdcut) {
                pdsum_sig += pdval;
                nelem_sig++;
            } else {
                pdsum_base += pdval;
                nelem_base++;
            }
        }
        double pdavg_sig = pdsum_sig  / nelem_sig;
        double pdavg_base = pdsum_base / nelem_base;
        System.out.println("Number elements in shutter open region = " + nelem_sig);
        // now integrate points in shutter open range
        double pdval_previous = 0.,
               pdtime_previous = 0.,
               pdtime_sum = 0.,
               frac_sig_previous = 0.;
        int    nint = 0;
        double pdint = 0.0; // SUM PD(t)*dt

        for (PDValue data : pddata) {
            double pdval = data.getValue();
            double pdtime = data.getTime();
            double frac_sig = (pdval-pdavg_base)/(pdavg_sig-pdavg_base);
            if (frac_sig>FRAC_SIG_MIN_LIMIT || frac_sig_previous>FRAC_SIG_MIN_LIMIT) { // use value if it or its previous value is at least FRAC_SIG_MIN_LIMIT sig level
                double delta_time = pdtime-pdtime_previous;
                // note that the following is corrected for a non-zero baseline
                pdint += ((pdval+pdval_previous)/2.0-pdavg_base)*delta_time;
                pdtime_sum += delta_time;
                nint++;
            }
            pdval_previous    = pdval;
            pdtime_previous   = pdtime;
            frac_sig_previous = frac_sig;
        }
                
        double pdavg_expo = pdint / pdtime_sum;

        System.out.println("Sum of PD values in luminous region = " + pdsum);
        System.out.println("PD average value during exposure = " + pdavg_sig);
        System.out.println("PD average value baseline = " + pdavg_base);
        System.out.println("PD average value sig-baseline = " + (pdavg_sig - pdavg_base));
        System.out.println("Number elements integrated = " + nint);
        System.out.println("Integration summed time = " + pdtime_sum);
        System.out.println("PD average from integration = " + pdavg_expo);
        System.out.println("NOTE: The MONDIODE value will use EXPTIME.");
        return pdint / 1.e-9; // convert to nA
    }

    private BinaryTableHDU createPhotoDiodeHDU(List<PDValue> data,String extnam, String c1name, String c2name, double tstart) throws FitsException {
        BinaryTable binTable = new BinaryTable();
        double[] times = new double[data.size()];
        double[] values = new double[data.size()];
        int i = 0;
        for (PDValue datum : data) {
            times[i] = datum.getTime();
            values[i] = datum.getValue();
            i++;
        }
        binTable.addColumn(times);
        binTable.addColumn(values);

        Header header = new Header(binTable);
        
        header.addValue("EXTNAME", extnam,"Name of the extension");
        header.addValue("TSTART", tstart, "Time of Start of Readings");

        final BinaryTableHDU bhdu = new BinaryTableHDU(header, binTable);
        bhdu.setColumnName(0, c1name, "time");
        bhdu.setColumnName(1, c2name, "values");
        bhdu.info(System.out);
        return bhdu;
    }

    /**
     * Shows the Photo Diode buffered readings analysis output without
     * modifying the corresponding fits file
     * 
     * @param pddatfile The file containing the photo diode data
     * @throws java.io.IOException
     * @throws nom.tam.fits.FitsException
     */
    public void showPhotoDiodeAnalysis(File pddatfile) throws IOException, FitsException {

        List<PDValue> data = readPhotoDiodeFile(pddatfile);
        double pdint = analyzePhotoDiodeValues(data);
    }
    /**
     * Update an existing files by adding or replacing the extension HDU
     * containing the binary table of photo diode values. Note this routine
     * currently works by copying the entire file to a temporary file, then
     * renaming the temporary file over the top of the original file.
     *
     * @param pddatfile The file containing the photo diode data
     * @param fitsfile The fits file to update
     * @param extnam The extension name to add/update
     * @param c1name The name for the first column of data
     * @param c2name The name for the second column of data
     * @param tstart The start time of the readings?
     * @throws java.io.IOException
     * @throws nom.tam.fits.FitsException
     */
    public void updatePhotoDiodeValues(File pddatfile, File fitsfile, String extnam, String c1name, String c2name, double tstart) throws IOException, FitsException {

        List<PDValue> data = readPhotoDiodeFile(pddatfile);
        double pdint = analyzePhotoDiodeValues(data);
        double pdavg = 0.; // guard value
        BinaryTableHDU bhdu = createPhotoDiodeHDU(data,extnam,c1name,c2name,tstart);

        File tmpName = File.createTempFile("tmp", "fits", fitsfile.getParentFile());
        try (Fits in = new Fits(fitsfile)) {
            try (BufferedFile out = new BufferedFile(tmpName, "rw")) {
                in.read();
                BasicHDU primary = in.getHDU(0);
                double exptime = primary.getHeader().getDoubleValue("EXPTIME",1.0);
                if (primary.getHeader().containsKey("MONDIODE")) {
                    primary.getHeader().deleteKey("MONDIODE");
                }
                pdavg = 0.;
                if (exptime>0.0) pdavg = pdint / exptime;
                primary.getHeader().addValue("MONDIODE", pdavg, "avg PD value (nA) during expo");
                FitsCheckSum.setChecksum(primary);
                primary.write(out);
                for (int ihdu = 1; ihdu < in.getNumberOfHDUs(); ihdu++) {
                    final BasicHDU hdu = in.getHDU(ihdu);
                    String extn = hdu.getHeader().getStringValue("EXTNAME");
                    if (!extn.equals(extnam)) {
                        hdu.write(out);
                    }
                }
                FitsCheckSum.setChecksum(bhdu);
                bhdu.write(out);
            }
        }
        Files.move(tmpName.toPath(), fitsfile.toPath(), StandardCopyOption.REPLACE_EXISTING);
    }

    /**
     * Compute the flux statistics by reading an existing FITS file
     *
     * @param fitsfile The file to read
     * @return The computed flux statistics.
     * @throws nom.tam.fits.FitsException If the fits file can not be read
     * @throws java.io.IOException If an IO exception occurs
     */
    public double getFluxStats(File fitsfile) throws FitsException, IOException {

        try (Fits fits = new Fits(fitsfile)) {
            double minflux = Double.MAX_VALUE;
            double fluxsum = 0.0;
                double[] fluxvals = new double[16];
            int fluxentries = 0;
            
            fits.read();
            final Header primary = fits.getHDU(0).getHeader();
            double gain = primary.getDoubleValue("CCDGAIN");
            double exptime = fits.getHDU(0).getHeader().getDoubleValue("EXPTIME");

            for (int ihdu = 1; ihdu < fits.getNumberOfHDUs(); ihdu++) {
                final Header extension = fits.getHDU(ihdu).getHeader();
                String extn = extension.getStringValue("EXTNAME");
                if (extn.contains("Segment")) {
                    double avg = fits.getHDU(ihdu).getHeader().getDoubleValue("AVERAGE"); // data region only
                    double bias = fits.getHDU(ihdu).getHeader().getDoubleValue("AVGBIAS"); // bias region only
                    double signal = (avg - bias) * gain;         // in electrons
                    double flux = Math.max(signal / exptime, 1.0);  // in electrons/second (can't be < 0 !)
                    minflux = Math.min(minflux, flux);
                    fluxsum += flux;
                    fluxvals[fluxentries] = flux;
                    fluxentries++;
                    System.out.println(extn + " signal = " + signal + " exptime = " + exptime + " flux = " + flux + " minflux for CCD = " + minflux);
                }
            }

            // ignore bad segments by only using flux values that are no less than a third of the average
            double flux_cutoff = fluxsum / (3.0 * fluxentries);

            int num_used = 0;
            double flux_resum = 0.0;
            for (int idx=0; idx<16 ; idx++) {
                if (fluxvals[idx]>flux_cutoff) {
                    flux_resum += fluxvals[idx];
                    num_used++;
                }
            }
            System.out.println("flux_cutoff = "+flux_cutoff+" , number flux values used = "+num_used);
            double sensor_flux = num_used > 0 ? flux_resum / num_used : 0.0 ;
//            double sensor_flux = (fluxsum - minflux) / (fluxentries-1);
            System.out.println("flux with all segments = "+fluxsum/fluxentries);
            System.out.println("flux with non-responsive segments removed = "+sensor_flux);
//            System.out.println("flux with min value removed = "+sensor_flux);
            return sensor_flux;
        }
    }
}
